"use strict";
var _a, _b;
Object.defineProperty(exports, "__esModule", { value: true });
exports.isSafeError = exports.SafeError = exports.isGrafastError = exports.newGrafastError = exports._GrafastError = exports.$$error = void 0;
const interfaces_js_1 = require("./interfaces.js");
exports.$$error = Symbol("isGrafastError");
// IMPORTANT: this WILL NOT WORK when compiled down to ES5. It requires ES6+
// native class support.
/**
 * When an error occurs during plan execution we wrap it in a GrafastError so
 * that we can pass it around as a value.  It gets unwrapped and thrown in the
 * grafast resolver.
 *
 * @internal
 */
const _GrafastError = class GrafastError extends Error {
    static { _a = exports.$$error; }
    constructor(originalError, planId) {
        if (originalError instanceof exports._GrafastError) {
            throw new Error("GrafastInternalError<62505509-8b21-4ef7-80f5-d0f99873174b>: attempted to wrap a GrafastError with a GrafastError.");
        }
        const message = originalError?.message;
        super(message);
        this[_a] = true;
        Object.setPrototypeOf(this, GrafastError.prototype);
        this.originalError = originalError;
        this.extensions = { grafast: { planId } };
    }
};
exports._GrafastError = _GrafastError;
/**
 * DO NOT ALLOW CONSTRUCTION OF ERRORS OUTSIDE OF THIS MODULE!
 *
 * @internal
 */
function newGrafastError(error, planId) {
    return new exports._GrafastError(error, planId);
}
exports.newGrafastError = newGrafastError;
/**
 * Is the given value a GrafastError? This is the only public API that people
 * should use for looking at GrafastErrors.
 */
function isGrafastError(value) {
    return typeof value === "object" && value !== null && exports.$$error in value;
}
exports.isGrafastError = isGrafastError;
class SafeError extends Error {
    static { _b = interfaces_js_1.$$safeError; }
    static { this.$$export = {
        moduleName: "grafast",
        exportName: "SafeError",
    }; }
    constructor(message, extensions = undefined, errorOptions) {
        super(message, errorOptions);
        this.extensions = extensions;
        this[_b] = true;
        Object.setPrototypeOf(this, SafeError.prototype);
    }
}
exports.SafeError = SafeError;
function isSafeError(error) {
    return error[interfaces_js_1.$$safeError];
}
exports.isSafeError = isSafeError;
//# sourceMappingURL=error.js.map