"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.applyTransforms = exports.ApplyTransformsStep = void 0;
const tslib_1 = require("tslib");
const assert = tslib_1.__importStar(require("../assert.js"));
const dev_js_1 = require("../dev.js");
const executeBucket_js_1 = require("../engine/executeBucket.js");
const LayerPlan_js_1 = require("../engine/LayerPlan.js");
const withGlobalLayerPlan_js_1 = require("../engine/lib/withGlobalLayerPlan.js");
const step_js_1 = require("../step.js");
const __item_js_1 = require("./__item.js");
/**
 * @internal
 */
class ApplyTransformsStep extends step_js_1.ExecutableStep {
    static { this.$$export = {
        moduleName: "grafast",
        exportName: "ApplyTransformsStep",
    }; }
    constructor(listPlan) {
        super();
        // OPTIMIZE: if all the steps in the inner bucket are synchronous then theoretically we can be synchronous too
        this.isSyncAndSafe = false;
        this.addDependency(listPlan);
        // Plan this subroutine
        this.subroutineLayer = new LayerPlan_js_1.LayerPlan(this.operationPlan, this.layerPlan, {
            type: "subroutine",
            parentStep: this,
        });
        const itemPlan = (0, withGlobalLayerPlan_js_1.withGlobalLayerPlan)(this.subroutineLayer, listPlan.polymorphicPaths, () => {
            // This does NOT use `itemPlanFor` because __ListTransformPlans are special.
            const $__listItem = new __item_js_1.__ItemStep(listPlan);
            $__listItem.transformStepId = this.id;
            this.itemStepId = $__listItem.id;
            const $listItem = (0, step_js_1.isListCapableStep)(listPlan)
                ? listPlan.listItem($__listItem)
                : $__listItem;
            const $newListItem = applyTransforms($listItem);
            if (this.isSyncAndSafe &&
                (!$__listItem.isSyncAndSafe ||
                    !$listItem.isSyncAndSafe ||
                    !$newListItem.isSyncAndSafe)) {
                this.isSyncAndSafe = false;
            }
            return $newListItem;
        });
        this.subroutineLayer.setRootStep(itemPlan);
        this.operationPlan.finishSubroutine(this, this.subroutineLayer);
    }
    async execute(_count, values, extra) {
        const bucket = extra._bucket;
        const childLayerPlan = this.subroutineLayer;
        const copyStepIds = childLayerPlan.copyStepIds;
        const store = new Map();
        const polymorphicPathList = [];
        const iterators = [];
        const map = new Map();
        let size = 0;
        // ENHANCE: do this better!
        const itemStepId = this.operationPlan.dangerouslyGetStep(this.itemStepId).id;
        if (itemStepId == null) {
            throw new Error("GrafastInternalError<b3a2bff9-15c6-47e2-aa82-19c862324f1a>: listItem layer plan has no rootStepId");
        }
        store.set(itemStepId, []);
        // Prepare store with an empty list for each copyPlanId
        for (const planId of copyStepIds) {
            store.set(planId, []);
            if (!bucket.store.has(planId)) {
                throw new Error(`GrafastInternalError<14f2b4c6-f951-44d6-ad6b-2eace3330b84>: plan '${planId}' (${this.operationPlan.dangerouslyGetStep(planId)}) listed in copyStepIds but not available in parent bucket for ${this}`);
            }
        }
        const listValues = values[0];
        // We'll typically be creating more listItem bucket entries than we
        // have parent buckets, so we must "multiply up" the store entries.
        for (let originalIndex = 0; originalIndex < listValues.length; originalIndex++) {
            const list = listValues[originalIndex];
            if (Array.isArray(list)) {
                const newIndexes = [];
                map.set(originalIndex, newIndexes);
                for (let j = 0, l = list.length; j < l; j++) {
                    const newIndex = size++;
                    newIndexes.push(newIndex);
                    polymorphicPathList[newIndex] =
                        bucket.polymorphicPathList[originalIndex];
                    // Copying across the iterators because we do NOT call outputBucket,
                    // so we need to ensure any streams are cleaned up.
                    iterators[newIndex] = bucket.iterators[originalIndex];
                    store.get(itemStepId)[newIndex] = list[j];
                    for (const planId of copyStepIds) {
                        store.get(planId)[newIndex] =
                            bucket.store.get(planId)[originalIndex];
                    }
                }
            }
        }
        if (size > 0) {
            const childBucket = (0, executeBucket_js_1.newBucket)({
                layerPlan: childLayerPlan,
                size,
                store,
                hasErrors: bucket.hasErrors,
                polymorphicPathList,
                iterators,
            }, bucket.metaByMetaKey);
            await (0, executeBucket_js_1.executeBucket)(childBucket, extra._requestContext);
        }
        const depResults = store.get(childLayerPlan.rootStep.id);
        return listValues.map((list, originalIndex) => {
            if (list == null) {
                return list;
            }
            const indexes = map.get(originalIndex);
            if (!Array.isArray(list) || !Array.isArray(indexes)) {
                // ERRORS: should this be an error?
                console.warn(`Either list or values was not an array when processing ${this}`);
                return null;
            }
            const values = indexes.map((idx) => {
                const val = depResults[idx];
                if (val instanceof Error) {
                    throw val;
                }
                return val;
            });
            if (dev_js_1.isDev) {
                assert.strictEqual(list.length, values.length, "GrafastInternalError<43cb302e-673b-4881-8c4c-f2d00fe5a3d7>: The list and values length must match for a ApplyTransformsStep");
            }
            return values;
        });
    }
}
exports.ApplyTransformsStep = ApplyTransformsStep;
// TODO: document this better. (Had to do it for `each($items, $item =>
// object({...}))` JSON transform in pg-many-to-many
/**
 * If you want to use a step that might represent a list and you need all of
 * the `listItem` transforms to have already taken place (e.g. you're going to
 * send the result to an external service) rather than processing them through
 * the GraphQL response, then you may need to call `applyTransforms` on it.
 */
function applyTransforms($step) {
    if ((0, step_js_1.isListCapableStep)($step)) {
        return new ApplyTransformsStep($step);
    }
    else {
        // No eval necessary
        return $step;
    }
}
exports.applyTransforms = applyTransforms;
//# sourceMappingURL=applyTransforms.js.map