"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.polymorphicBranch = exports.PolymorphicBranchStep = void 0;
const index_js_1 = require("../index.js");
const step_js_1 = require("../step.js");
const constant_js_1 = require("./constant.js");
class PolymorphicBranchStep extends step_js_1.ExecutableStep {
    static { this.$$export = {
        moduleName: "grafast",
        exportName: "PolymorphicBranchStep",
    }; }
    constructor($step, matchers) {
        super();
        this.isSyncAndSafe = true;
        this.addDependency($step);
        this.typeNames = Object.keys(matchers);
        this.matchers = Object.fromEntries(Object.entries(matchers).map(([typeName, matcher]) => {
            const fixedMatcher = {
                match: matcher.match ?? ((obj) => obj.__typename === typeName),
                plan: matcher.plan ?? (($obj) => $obj),
            };
            return [typeName, fixedMatcher];
        }));
    }
    planForType(objectType) {
        const matcher = this.matchers[objectType.name];
        const $step = this.getDep(0);
        if (matcher) {
            if (typeof matcher.plan === "function") {
                return matcher.plan($step);
            }
            else {
                return $step;
            }
        }
        else {
            // TODO: should this be an error?
            return (0, constant_js_1.constant)(null);
        }
    }
    execute(count, values, _extra) {
        const results = [];
        const objects = values[0];
        for (let i = 0; i < count; i++) {
            const obj = objects[i];
            let match = null;
            if (obj != null) {
                for (const typeName of this.typeNames) {
                    if (this.matchers[typeName].match(obj)) {
                        match = typeName;
                        break;
                    }
                }
            }
            results[i] = match !== null ? (0, index_js_1.polymorphicWrap)(match, obj) : null;
        }
        return results;
    }
}
exports.PolymorphicBranchStep = PolymorphicBranchStep;
function polymorphicBranch($step, matchers) {
    return new PolymorphicBranchStep($step, matchers);
}
exports.polymorphicBranch = polymorphicBranch;
//# sourceMappingURL=polymorphicBranch.js.map